# The LearningOnline Network with CAPA
# Routines for messaging display
#
# $Id: lonmsgdisplay.pm,v 1.51 2006/12/08 20:32:36 raeburn Exp $
#
# Copyright Michigan State University Board of Trustees
#
# This file is part of the LearningOnline Network with CAPA (LON-CAPA).
#
# LON-CAPA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LON-CAPA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LON-CAPA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# /home/httpd/html/adm/gpl.txt
#
# http://www.lon-capa.org/
#


package Apache::lonmsgdisplay;

=pod

=head1 NAME

Apache::lonmsg: supports internal messaging

=head1 SYNOPSIS

lonmsg provides routines for sending messages, receiving messages, and
a handler to allow users to read, send, and delete messages.

=head1 OVERVIEW

=head2 Messaging Overview

X<messages>LON-CAPA provides an internal messaging system similar to
email, but customized for LON-CAPA's usage. LON-CAPA implements its
own messaging system, rather then building on top of email, because of
the features LON-CAPA messages can offer that conventional e-mail can
not:

=over 4

=item * B<Critical messages>: A message the recipient B<must>
acknowlegde receipt of before they are allowed to continue using the
system, preventing a user from claiming they never got a message

=item * B<Receipts>: LON-CAPA can reliably send reciepts informing the
sender that it has been read; again, useful for preventing students
from claiming they did not see a message. (While conventional e-mail
has some reciept support, it's sporadic, e-mail client-specific, and
generally the receiver can opt to not send one, making it useless in
this case.)

=item * B<Context>: LON-CAPA knows about the sender, such as where
they are in a course. When a student mails an instructor asking for
help on the problem, the instructor receives not just the student's
question, but all submissions the student has made up to that point,
the user's rendering of the problem, and the complete view the student
saw of the resource, including discussion up to that point. Finally,
the instructor is reading all of this inside of LON-CAPA, not their
email program, so they have full access to LON-CAPA's grading
interface, or other features they may wish to use in response to the
student's query.

=item * B<Blocking>: LON-CAPA can block selected communication 
features for students during an online exam. A course coordinator or
instructor can set an open and close date/time for scheduled online
exams in a course. If a user uses the LON-CAPA internal messaging 
system to display e-mails during the scheduled blocking event,  
display of all e-mail sent during the blocking period will be 
suppressed, and a message of explanation, including details of the 
currently active blocking periods will be displayed instead. A user 
who has a course coordinator or instructor role in a course will be
unaffected by any blocking periods for the course, unless the user
also has a student role in the course, AND has selected the student role.

=back

Users can ask LON-CAPA to forward messages to conventional e-mail
addresses on their B<PREF> screen, but generally, LON-CAPA messages
are much more useful than traditional email can be made to be, even
with HTML support.

Right now, this document will cover just how to send a message, since
it is likely you will not need to programmatically read messages,
since lonmsg already implements that functionality.

The routines used to package messages and unpackage messages are not
only used by lonmsg when creating/extracting messages for LON-CAPA's
internal messaging system, but also by lonnotify.pm which is available
for use by Domain Coordinators to broadcast standard e-mail to specified
users in their domain.  The XML packaging used in the two cases is very
similar.  The differences are the use of <recuser>$uname</recuser> and 
<recdomain>$udom</recdomain> in stored internal messages, compared 
with <recipient username="$uname:$udom">$email</recipient> in stored
Domain Coordinator e-mail for the storage of information about 
recipients of the message/e-mail.

=head1 FUNCTIONS

=over 4

=cut

use strict;
use Apache::lonnet;
use HTML::TokeParser();
use Apache::Constants qw(:common);
use Apache::loncommon();
use Apache::lontexconvert();
use HTML::Entities();
use Apache::lonlocal;
use Apache::loncommunicate;
use Apache::lonfeedback;
use Apache::lonrss();
use Apache::lonselstudent();
use lib '/home/httpd/lib/perl/';
use LONCAPA;

# Querystring component with sorting type
my $sqs;
my $startdis;
my $interdis;

# ============================================================ List all folders

sub folderlist {
    my $folder=shift;
    my %lt = &Apache::lonlocal::texthash(
                actn => 'Action',
                fold => 'Folder',
                show => 'Show',
                go   => 'Go',
                nnff => 'New Name for Folder',
                newn => 'New Name',
                thfm => 'The folder may not be renamed',
                fmnb => 'folder may not be renamed as it is a folder provided by the system.',
                asth => 'as this name is already in use for a system-provided or user-defined folder.',
                the => 'The',
                tnfm => 'The new folder may not be named',

    );

    my %actions = &Apache::lonlocal::texthash(
                                view => 'View Folder',
                                rename => 'Rename Folder',
                                delete => 'Delete Folder',
    );
    $actions{'select_form_order'} = ['view','rename','delete'];

    my %permfolders = &get_permanent_folders();
    my $permlist = join("','",sort(keys(%permfolders)));
    my ($permlistkeys,$permlistvals);
    foreach my $key (sort(keys(%permfolders))) {
        $permlistvals .= $permfolders{$key}."','";
        $permlistkeys .= $key."','";
    }
    $permlistvals =~ s/','$//;
    $permlistkeys =~ s/','$//;
    my %gotfolders = &Apache::lonmsg::get_user_folders();
    my %userfolders;

    foreach my $key (keys(%gotfolders)) {
        $userfolders{$key} = $key;
    }
    my @userorder = sort(keys(%userfolders));
    my %formhash = (%permfolders,%userfolders);
    my $folderlist = join("','",@userorder);
    $folderlist .= "','".$permlistvals;

    $formhash{'select_form_order'} = ['','critical','new',@userorder,'sent','trash'];
    my $output = qq|<script type="text/javascript">
function folder_choice(targetform,caller) {
    var permfolders_keys = new Array('$permlistkeys');
    var permfolders_vals = new Array('$permlistvals');
    var allfolders = new Array('$folderlist');
    if (caller == 'change') {
        if (targetform.folderaction.options[targetform.folderaction.selectedIndex].value == 'rename') {
            for (var i=0; i<permfolders_keys.length; i++) {
                if (permfolders_keys[i] == targetform.folder.value) {
                    alert("$lt{'the'} '"+permfolders_vals[i]+"' $lt{'fmnb'}");
                    return;
                }
            }
            var foldername=prompt('$lt{'nnff'}','$lt{'newn'}');
            if (foldername) {
                targetform.renamed.value=foldername;
                for (var i=0; i<allfolders.length; i++) {
                    if (allfolders[i] == foldername) {
                        alert("$lt{'thfm'} '"+foldername+"' $lt{'asth'}");
                        return;
                    }
                }
                targetform.submit();
            }
        }
        else {
            targetform.submit();
        }
    }
    if (caller == 'new') {
        var newname=targetform.newfolder.value;
        if (newname) {
            for (var i=0; i<allfolders.length; i++) {
                if (allfolders[i] == newname) {
                    alert("$lt{'tnfm'} '"+newname+"' $lt{'asth'}");
                    return;
                }
            }
            targetform.submit();
        }
    }
}
</script>|;
    $output .= '
<form method="post" action="/adm/email" name="folderlist">
<table border="0" cellspacing="2" cellpadding="2">
 <tr>
  <td align="left">
   <table border="0" cellspacing="2" cellpadding="2">
    <tr>
     <td align="center"><b>'.$lt{'fold'}.'</b><br />'."\n".
         &Apache::loncommon::select_form($folder,'folder',%formhash).'
     </td>
     <td align="center"><b>'.$lt{'show'}.'</b><br />
      <select name="interdis">'.
          join("\n",map { '<option value="'.$_.'"'.
			      ($_==$interdis?' selected="selected"':'').'>'.$_.'</option>' 
			}
	       (10,20,50,100,200)).'</select>
     </td>
     <td align="center"><b>'.$lt{'actn'}.'</b><br />'.
         &Apache::loncommon::select_form('view','folderaction',%actions).'
     </td><td><br />'.
    '<input type="button" value="'.$lt{'go'}.'" onClick="javascript:folder_choice(this.form,'."'change'".');" />
     </td>
    </tr>
   </table>
  </td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td>
  <td align="right">
   <table><tr><td><br />
    <input type="button" value="'.&mt('Make New Folder').
    '" onClick="javascript:folder_choice(this.form,'."'new'".');" /></td>'.
    '<td align="center"><b>'.&mt('New Folder').'</b><br />'.
    '<input type="text" size="15" name="newfolder" value="" />
    </td></tr></table>
  </td>
 </tr>
</table>'."\n".
    '<input type="hidden" name="sortedby" value="'.$env{'form.sortedby'}.'" />'.
    '<input type="hidden" name="renamed" value="" />'.
        ($folder=~/^(new|critical)/?'</form>':'');
    return $output;
}

sub get_permanent_folders {
    my %permfolders = 
	&Apache::lonlocal::texthash(''         => 'INBOX',
				    'trash'    => 'TRASH',
				    'new'      => 'New Messages Only',
				    'critical' => 'Critical',
				    'sent'     => 'Sent Messages',
				    );
    return %permfolders;
}

sub scrollbuttons {
    my ($start,$maxdis,$first,$finish,$total)=@_;
    unless ($total>0) { return ''; }
    $start++; $maxdis++;$first++;$finish++;
    return
   &mt('Page').': '. 
   '<input type="submit" name="firstview" value="'.&mt('First').'" />'.
   '<input type="submit" name="prevview" value="'.&mt('Previous').'" />'.
   '<input type="text" size="5" name="startdis" value="'.$start.'" onChange="this.form.submit()" /> of '.$maxdis.
   '<input type="submit" name="nextview" value="'.&mt('Next').'" />'.
   '<input type="submit" name="lastview" value="'.&mt('Last').'" /><br />'.
   &mt('Showing messages [_1] through [_2] of [_3]',$first,$finish,$total).'</form>';
}
# =============================================================== Status Change

sub statuschange {
    my ($msgid,$newstatus,$folder)=@_;
    my $suffix=&Apache::lonmsg::foldersuffix($folder);
    my %status=&Apache::lonnet::get('email_status'.$suffix,[$msgid]);
    if ($status{$msgid}=~/^error\:/) { $status{$msgid}=''; }
    unless ($status{$msgid}) { $status{$msgid}='new'; }
    unless (($status{$msgid} eq 'replied') || 
            ($status{$msgid} eq 'forwarded')) {
	&Apache::lonnet::put('email_status'.$suffix,{$msgid => $newstatus});
    }
    if (($newstatus eq 'deleted') || ($newstatus eq 'new')) {
	&Apache::lonnet::put('email_status'.$suffix,{$msgid => $newstatus});
    }
    if ($newstatus eq 'deleted') {
	return &movemsg($msgid,$folder,'trash');
    }
    return ;
}

# ============================================================= Make new folder

sub makefolder {
    my ($newfolder) = @_;
    my %permfolders = &get_permanent_folders();
    my %userfolders = &Apache::lonmsg::get_user_folders();
    my ($outcome,$warning);
    if (defined($userfolders{$newfolder})) {
        return &mt('The folder name: "[_1]" is already in use for an existing folder.',$newfolder);
    }
    foreach my $perm (keys(%permfolders)) {
        if ($permfolders{$perm} eq $newfolder) {
            return &mt('The folder name: "[_1]" is already used for one of the folders automatically generated by the system.',$newfolder);
        }
    } 
    if (&get_msgfolder_lock() eq 'ok') {
        my %counter_hash = &Apache::lonnet::get('email_folders',["\0".'idcount']);
        my $lastcount = $counter_hash{"\0".'idcount'};
        my $folder_id = $lastcount + 1;
        while (defined($userfolders{$folder_id})) {
            $folder_id ++;
        }
        my %folderinfo = ( id      => $folder_id,
                           created => time, );
        $outcome =  
	    &Apache::lonnet::put('email_folders',{$newfolder => \%folderinfo,
						  "\0".'idcount' => $folder_id});
        my $releaseresult = &release_msgfolder_lock();
        if ($releaseresult ne 'ok') {
            $warning = $releaseresult;
        }
    } else {
        $outcome = 
	    &mt('Error - could not obtain lock on email folders record.');
    }
    return ($outcome,$warning);
}

# ============================================================= Delete folder

sub deletefolder {
    my ($folder)=@_;
    my %permfolders = &get_permanent_folders();
    if (defined($permfolders{$folder})) {
        return &mt('The folder [_1] may not be deleted',$folder); 
    }
    my %userfolders = &Apache::lonmsg::get_user_folders();
    if (!defined($userfolders{$folder})) {
        return &mt('The folder [_1] does not exist so deletion is not required.',
                   $folder);
    }
    # check folder is empty;
    my $suffix=&Apache::lonmsg::foldersuffix($folder);
    my @messages = &Apache::lonnet::getkeys('nohist_email'.$suffix);
    if (@messages > 0) {
        return &mt('The folder [_1] contains messages so it may not be deleted.').
               '<br />'.
               &mt('Delete or move the messages to a different folder first.');
    }
    my $delresult = &Apache::lonnet::del('email_folders',[$folder]);
    return $delresult;
}

sub renamefolder {
    my ($folder) = @_;
    my $newname = $env{'form.renamed'};
    my %permfolders = &get_permanent_folders();
    if ($env{'form.renamed'} eq '') {
        return &mt('The folder "[_1]" may not be renamed to "[_2]" as the new name you requested is an invalid name.',$folder,$newname);
    }
    if (defined($permfolders{$newname})) {
        return &mt('The folder "[_1]" may not be renamed to "[_2]" as the new name you requested is reserved for folders provided automatically by the system.',$folder,$newname);
    }
    my %userfolders = &Apache::lonmsg::get_user_folders();
    if (defined($userfolders{$newname})) {
        return &mt('The folder "[_1]" may not be renamed to "[_2]" because the new name you requested is already being used for an existing folder.',$folder,$newname);
    }
    if (!defined($userfolders{$folder})) {
        return &mt('The folder "[_1]" could not be renamed to "[_2]" because the folder does not exist.',$folder,$newname);
    }
    my %folderinfo;
    if (ref($userfolders{$folder}) eq 'HASH') {
        %folderinfo = %{$userfolders{$folder}};
    } else {
        %folderinfo = ( id      => $folder,
                        created => $userfolders{$folder},);
    }
    my $outcome =
     &Apache::lonnet::put('email_folders',{$newname => \%folderinfo});
    if ($outcome eq 'ok') {
        $outcome = &Apache::lonnet::del('email_folders',[$folder]);
    }
    return $outcome;
}

sub get_msgfolder_lock {
    # get lock for mail folder counter.
    my $lockhash = { "\0".'lock_counter' => time, };
    my $tries = 0;
    my $gotlock = &Apache::lonnet::newput('email_folders',$lockhash);
    while (($gotlock ne 'ok') && $tries <3) {
        $tries ++;
        sleep(1);
        $gotlock = &Apache::lonnet::newput('email_folders',$lockhash);
    }
    return $gotlock;
}

sub release_msgfolder_lock {
    #  remove lock
    my @del_lock = ("\0".'lock_counter');
    my $dellockoutcome=&Apache::lonnet::del('email_folders',\@del_lock);
    if ($dellockoutcome ne 'ok') {
        return ('<br />'.&mt('Warning: failed to release lock for counter').'<br />');
    } else {
        return 'ok';
    }
}


# ======================================================== Move between folders

sub movemsg {
    my ($msgid,$srcfolder,$trgfolder)=@_;
    if ($srcfolder eq 'new') { $srcfolder=''; }
    my $srcsuffix=&Apache::lonmsg::foldersuffix($srcfolder);
    my $trgsuffix=&Apache::lonmsg::foldersuffix($trgfolder);
    if ($srcsuffix eq $trgsuffix) {
	return (0,&mt('Message not moved, Attempted to move message to the same folder as it already is in.'));
    }

# Copy message
    my %message=&Apache::lonnet::get('nohist_email'.$srcsuffix,[$msgid]);
    if (!exists($message{$msgid}) || $message{$msgid} eq '') {
	if (&Apache::lonnet::error(%message)) {
	    return (0,&mt('Message not moved, A network error occurred.'));
	} else {
	    return (0,&mt('Message not moved as the message is no longer in the source folder.'));
	}
    }

    my $result =&Apache::lonnet::put('nohist_email'.$trgsuffix,
				     {$msgid => $message{$msgid}});
    if (&Apache::lonnet::error($result)) {
	return (0,&mt('Message not moved, A network error occurred.'));
    }

# Copy status
    unless ($trgfolder eq 'trash') {
       	my %status=&Apache::lonnet::get('email_status'.$srcsuffix,[$msgid]);
	# a non-existant status is the mark of an unread msg
	if (&Apache::lonnet::error(%status)) {
	    return (0,&mt('Message copied to new folder but status was not, A network error occurred.'));
	}
	my $result=&Apache::lonnet::put('email_status'.$trgsuffix,
					{$msgid => $status{$msgid}});
	if (&Apache::lonnet::error($result)) {
	    return (0,&mt('Message copied to new folder but status was not, A network error occurred.'));
	}
    }

# Delete orginals
    my $result_del_msg = 
	&Apache::lonnet::del('nohist_email'.$srcsuffix,[$msgid]);
    my $result_del_stat =
	&Apache::lonnet::del('email_status'.$srcsuffix,[$msgid]);
    if (&Apache::lonnet::error($result_del_msg)) {
	return (0,&mt('Message copied, but unable to delete the original from the source folder.'));
    }
    if (&Apache::lonnet::error($result_del_stat)) {
	return (0,&mt('Message copied, but unable to delete the original status from the source folder.'));
    }

    return (1);
}

# ======================================================= Display a course list

sub discourse {
    my $result;
    my ($course_personnel,
	$current_members,
	$expired_members,
	$future_members) = 
	    &Apache::lonselstudent::get_people_in_class($env{'request.course.sec'});
    unshift @$current_members, (@$course_personnel);
    my %defaultUsers;
    
    $result .= '<input type="hidden" name="sendmode" value="group" />'."\n";

    $result .= &Apache::lonselstudent::render_student_list($current_members,
							   "compemail",
							   "current",
							   \%defaultUsers,
							   1,"selectedusers",1);

    $result .= &Apache::lonselstudent::render_student_list($expired_members,
							   "compemail",
							   "expired",
							   \%defaultUsers,
							   1, "selectedusers",0);
    $result .= &Apache::lonselstudent::render_student_list($future_members,
							   "compemail",
							   "future",
							   \%defaultUsers,
							   1, "selectedusers", 0);
    return $result;
}

sub disgroup {
    my ($cdom,$cnum,$group,$viewgrps,$editgrps) = @_;
    my $result;
    #  Needs to be in a course
    if (!($env{'request.course.fn'})) {
        $result = &mt('Error: you must have a course role selected to be able to send a broadcast message to a group in the course.');
        return $result;
    }
    if ($cdom eq '' || $cnum eq '') {
        $result = &mt('Error: could not determine domain or number of course');
        return $result;
    }
    my ($memberinfo,$numitems) =
                 &Apache::longroup::group_memberlist($cdom,$cnum,$group,{},[]);
    my @statustypes = ('active');
    if ($viewgrps || $editgrps) {
        push(@statustypes,('future','previous'));
    }
    if (keys(%{$memberinfo}) == 0) {
        $result = &mt('As this group has no members, there are no '.
                      'recipients to select.');
        return $result;
    } else {
        $result = &mt('Select message recipients from the group members listed below.<br />');  
        my %Sortby = (
                         active   => {},
                         previous => {},
                         future   => {},
                     );
        my %lt = &Apache::lonlocal::texthash(
                                     'name'     => 'Name',
                                     'usnm'     => 'Username',
                                     'doma'     => 'Domain',
                                     'active'   => 'Active Members',
                                     'previous' => 'Former Members',
                                     'future'   => 'Future Members',
                                    );
        foreach my $user (sort(keys(%{$memberinfo}))) {
            my $status = $$memberinfo{$user}{status};
            if ($env{'form.'.$status.'.sortby'} eq 'fullname') {
                push(@{$Sortby{$status}{$$memberinfo{$user}{fullname}}},$user);
            } elsif ($env{'form.'.$status.'.sortby'} eq 'username') {
                push(@{$Sortby{$status}{$$memberinfo{$user}{uname}}},$user);
            } elsif ($env{'form.'.$status.'.sortby'} eq 'domain') {
                push(@{$Sortby{$status}{$$memberinfo{$user}{udom}}},$user);
            } else {
                push(@{$Sortby{$status}{$$memberinfo{$user}{fullname}}},$user);
            }
        }
        $result .= &group_check_uncheck();
        $result .= '<table border="0" cellspacing="8" cellpadding="2">'.
                   '<tr>';
        foreach my $status (@statustypes)  {
            if (ref($numitems) eq 'HASH') {
                if ((defined($$numitems{$status})) && ($$numitems{$status})) {
                    $result.='<td valign="top">'.
                             '<fieldset><legend><b>'.$lt{$status}.
                             '</b></legend><nobr>'.
                             '<input type="button" value="'.&mt('Check All').'" '.
                             'onclick="javascript:toggleAll('."'".$status."','check'".')" />'.
                             '&nbsp;&nbsp;'.
                             '<input type="button" value="'.&mt('Uncheck All').'" '.
                             'onclick="javascript:toggleAll('."'".$status."','uncheck'".')" />'.
                             '</nobr></fieldset><br />'.
                             &Apache::loncommon::start_data_table().
                             &Apache::loncommon::start_data_table_header_row();
                    $result .= "<th>$lt{'name'}</a></th>".
                               "<th>$lt{'usnm'}</a></th>".
                               "<th>$lt{'doma'}</a></th>".
                    &Apache::loncommon::end_data_table_header_row();
                    foreach my $key (sort(keys(%{$Sortby{$status}}))) {
                        foreach my $user (@{$Sortby{$status}{$key}}) {
                            $result .=
                                &Apache::loncommon::start_data_table_row().
                                '<td><input type="checkbox" '.
                                'name="selectedusers_forminput" value="'.
                                $user.':'.$status.'" />'.
                                $$memberinfo{$user}{'fullname'}.'</td>'.
                                '<td>'.$$memberinfo{$user}{'uname'}.'</td>'.
                                '<td>'.$$memberinfo{$user}{'udom'}.'</td>'.
                                &Apache::loncommon::end_data_table_row();
                        }
                    }
                    $result .= &Apache::loncommon::end_data_table();
                }
            }
            $result .= '</td><td>&nbsp;&nbsp;</td>';
        }
        $result .= '</tr></table>';
    }
    return $result;
}

sub group_check_uncheck {
    my $output = qq|
<script type="text/javascript">
function toggleAll(caller,action) {
    var pattern = new RegExp(":"+caller+"\$");
    if (typeof(document.compemail.selectedusers_forminput.length)=="undefined") {
        if (document.compemail.selectedusers_forminput.value.match(pattern)) {
            if (action == 'check') {
                document.groupmail.selectedusers_forminput.checked = true;
            } else {
                document.groupmail.selectedusers_forminput.checked = false;
            }
        }
    } else {
        for (var i=0; i<document.compemail.selectedusers_forminput.length; i++) {
            if (document.compemail.selectedusers_forminput[i].value.match(pattern)) {
                if (action == 'check') {
                    document.compemail.selectedusers_forminput[i].checked = true;
                } else {
                    document.compemail.selectedusers_forminput[i].checked = false;
                }
            }
        }
    }
}
</script>
    |;
}

sub groupmail_header {
    my ($action,$group,$cdom,$cnum) = @_;
    my ($description,$refarg);
    if (!$cdom || !$cnum) {
        $cdom = $env{'course.'.$env{'request.course.id'}.'.domain'};
        $cnum = $env{'course.'.$env{'request.course.id'}.'.num'};
    }
    if (exists($env{'form.ref'})) {
        $refarg = 'ref='.$env{'form.ref'};
    }
    if (!$group) {
        $group = $env{'form.group'};
    }
    if ($group eq '') {
        return  '';
    } else {
        my %curr_groups = &Apache::longroup::coursegroups($cdom,$cnum,$group);
        if (defined($curr_groups{$group})) {
            my %groupinfo =
                    &Apache::longroup::get_group_settings($curr_groups{$group});
            $description = &unescape($groupinfo{'description'});
        }
    }
    &Apache::lonhtmlcommon::clear_breadcrumbs();
    if ($refarg) {
        &Apache::lonhtmlcommon::add_breadcrumb
            ({href=>"/adm/coursegroups",
              text=>"Groups",
              title=>"View course groups"});
    }
    &Apache::lonhtmlcommon::add_breadcrumb
        ({href=>"/adm/$cdom/$cnum/$group/smppg?$refarg",
          text=>"Group: $description",
          title=>"Go to group's home page"},
         {href=>"/adm/email?compose=group&amp;group=".
                "$env{'form.group'}&amp;$refarg",
          text=>"Send a Message in a Group",
          title=>"Compose Group Email Message"},);
    if ($action eq 'sending') {
            &Apache::lonhtmlcommon::add_breadcrumb
                         ({text=>"Messages being sent.",
                           title=>"Messages sent"},);
    }
    my $groupheader = &Apache::loncommon::start_page('Group Email');
    $groupheader .= &Apache::lonhtmlcommon::breadcrumbs
                ('Group - '.$env{'form.group'}.' Email');
    return $groupheader;
}

sub groupmail_sent {
    my ($group,$cdom,$cnum) = @_;
    my $refarg;
    if (exists($env{'form.ref'})) {
        $refarg = 'ref='.$env{'form.ref'};
    }
    my $output .= '<br /><br /><a href="/adm/email?compose=group&amp;group='.
                  $group.'&amp;'.$refarg.'">'.
                  &mt('Send another group email').'</a>'.'&nbsp;&nbsp;&nbsp;'.
                  '<a href="/adm/'.$cdom.'/'.$cnum.'/'.$group.'/smppg?'.
                  $refarg.'">'. &mt('Return to group page').'</a>';
    return $output;
}

# ==================================================== Display Critical Message

sub discrit {
    my $r=shift;
    my $header = '<h1><font color="red">'.&mt('Critical Messages').'</font></h1>'.
        '<form action="/adm/email" method="POST">'.
        '<input type="hidden" name="confirm" value="true" />';
    my %what=&Apache::lonnet::dump('critical');
    my $result = '';
    foreach my $key (sort(keys(%what))) {
        my %content=&Apache::lonmsg::unpackagemsg($what{$key});
        next if ($content{'senderdomain'} eq '');
        $result.='<hr />'.&mt('From').': <b>'.
&Apache::loncommon::aboutmewrapper(
 &Apache::loncommon::plainname($content{'sendername'},$content{'senderdomain'}),$content{'sendername'},$content{'senderdomain'}).'</b> ('.
$content{'sendername'}.':'.
            $content{'senderdomain'}.') '.$content{'time'}.
            '<br />'.&mt('Subject').': '.$content{'subject'}.
            '<br /><pre>'.
              &Apache::lontexconvert::msgtexconverted($content{'message'}).
            '</pre><small>'.
&mt('You have to confirm that you received this message. After confirmation, this message will be moved to your regular inbox').
            '</small><br />'.
            '<input type="submit" name="rec_'.$key.'" value="'.&mt('Confirm Receipt').'" />'.
            '<input type="submit" name="reprec_'.$key.'" '.
                  'value="'.&mt('Confirm Receipt and Reply').'" />';
    }
    # Check to see if there were any messages.
    if ($result eq '') {
        $result = "<h2>".&mt('You have no critical messages.')."</h2>".
	    '<a href="/adm/roles">'.&mt('Select a course').'</a><br />'.
            '<a href="/adm/email">'.&mt('Communicate').'</a>';
    } else {
        $r->print($header);
    }
    $r->print($result);
    $r->print('<input type="hidden" name="displayedcrit" value="true" /></form>');
}

sub sortedmessages {
    my ($blocked,$startblock,$endblock,$numblocked,$folder) = @_;
    my $suffix=&Apache::lonmsg::foldersuffix($folder);
    my @messages = &Apache::lonnet::getkeys('nohist_email'.$suffix);

    #unpack the varibles and repack into temp for sorting
    my @temp;
    my %descriptions;
    my %status_cache = 
	&Apache::lonnet::get('email_status'.&Apache::lonmsg::foldersuffix($folder),\@messages);

    my $get_received;
    if ($folder eq 'sent' 
	&& ($env{'form.sortedby'} =~ m/^(rev)?(user|domain)$/)) {
	$get_received = 1;
    }

    foreach my $msgid (@messages) {
	my $esc_msgid=&escape($msgid);
	my ($sendtime,$shortsubj,$fromname,$fromdomain,$status,$fromcid)=
	    &Apache::lonmsg::unpackmsgid($esc_msgid,$folder,undef,
					 \%status_cache);
        my $description = &get_course_desc($fromcid,\%descriptions);
	my @temp1 = ($sendtime,$shortsubj,$fromname,$fromdomain,$status,
		     $esc_msgid,$description);
	if ($get_received) {
	    my %message = &Apache::lonnet::get('nohist_email'.$suffix,
					       [$msgid]);
	    my %content = &Apache::lonmsg::unpackagemsg($message{$msgid});
	    push(@temp1,$content{'recuser'},$content{'recdomain'});
	}
        # Check whether message was sent during blocking period.
        if ($sendtime >= $startblock && ($sendtime <= $endblock && $endblock > 0) ) {
            $$blocked{$msgid} = 'ON';
            $$numblocked ++;
        } else { 
            push @temp ,\@temp1;
        }
    }
    #default sort
    @temp = sort  {$a->[0] <=> $b->[0]} @temp;    
    if ($env{'form.sortedby'} eq "date"){
        @temp = sort  {$a->[0] <=> $b->[0]} @temp;    
    }
    if ($env{'form.sortedby'} eq "revdate"){
    	@temp = sort  {$b->[0] <=> $a->[0]} @temp; 
    }
    if ($env{'form.sortedby'} eq "user"){
	if ($get_received) {
	    @temp = sort  {lc($a->[7][0]) cmp lc($b->[7][0])} @temp;
	} else {
	    @temp = sort  {lc($a->[2])    cmp lc($b->[2])}    @temp;
	}
    }
    if ($env{'form.sortedby'} eq "revuser"){
	if ($get_received) {
	    @temp = sort  {lc($b->[7][0]) cmp lc($a->[7][0])} @temp;
	} else {
	    @temp = sort  {lc($b->[2])    cmp lc($a->[2])}    @temp;
	}
    }
    if ($env{'form.sortedby'} eq "domain"){
	if ($get_received) {
	    @temp = sort  {$a->[8][0] cmp $b->[8][0]} @temp;
	} else {
	    @temp = sort  {$a->[3]    cmp $b->[3]}    @temp;
	}
    }
    if ($env{'form.sortedby'} eq "revdomain"){
	if ($get_received) {
	    @temp = sort  {$b->[8][0] cmp $a->[8][0]} @temp;
	} else {
	    @temp = sort  {$b->[3]    cmp $a->[3]}    @temp;
	}
    }
    if ($env{'form.sortedby'} eq "subject"){
        @temp = sort  {lc($a->[1]) cmp lc($b->[1])} @temp;
    }
    if ($env{'form.sortedby'} eq "revsubject"){
        @temp = sort  {lc($b->[1]) cmp lc($a->[1])} @temp;
    }
    if ($env{'form.sortedby'} eq "course"){
        @temp = sort  {lc($a->[6]) cmp lc($b->[6])} @temp;
    }
    if ($env{'form.sortedby'} eq "revcourse"){
        @temp = sort  {lc($b->[6]) cmp lc($a->[6])} @temp;
    }
    if ($env{'form.sortedby'} eq "status"){
        @temp = sort  {$a->[4] cmp $b->[4]} @temp;
    }
    if ($env{'form.sortedby'} eq "revstatus"){
        @temp = sort  {$b->[4] cmp $a->[4]} @temp;
    }
    return @temp;
}

sub get_course_desc {
    my ($fromcid,$descriptions) = @_;
    my $description;
    if (!$fromcid) {
        return $description;
    } else {
        if (defined($$descriptions{$fromcid})) {
            $description = $$descriptions{$fromcid};
        } else {
            if (defined($env{'course.'.$fromcid.'.description'})) {
                $description = $env{'course.'.$fromcid.'.description'};
            } else {
                my %courseinfo=&Apache::lonnet::coursedescription($fromcid);
                $description = $courseinfo{'description'};
            }
            $$descriptions{$fromcid} = $description;
        }
        return $description;
    }
}

# ======================================================== Display new messages


sub disnew {
    my $r=shift;
    my %lt=&Apache::lonlocal::texthash(
				       'nm' => 'New Messages',
				       'su' => 'Subject',
                                       'co' => 'Course',
				       'da' => 'Date',
				       'us' => 'Username',
				       'op' => 'Open',
				       'do' => 'Domain'
				       );
    my @msgids = sort(&Apache::lonnet::getkeys('nohist_email'));
    my @newmsgs;
    my %setters = ();
    my %blocked = ();
    my $numblocked = 0;
    # Check for blocking of display because of scheduled online exams.
    my ($startblock,$endblock) = &Apache::loncommon::blockcheck(\%setters,'com');
    my %status_cache = 
	&Apache::lonnet::get('email_status',\@msgids);
    my %descriptions;
    foreach my $id (@msgids) {
	my $msgid=&escape($id);
        my ($sendtime,$shortsubj,$fromname,$fromdom,$status,$fromcid)=
	    &Apache::lonmsg::unpackmsgid($msgid,undef,undef,\%status_cache);
        if (defined($sendtime) && $sendtime!~/error/) {
            my $description = &get_course_desc($fromcid,\%descriptions);
            my $numsendtime = $sendtime;
            $sendtime = &Apache::lonlocal::locallocaltime($sendtime);
            if ($status eq 'new') {
                if ($numsendtime >= $startblock && ($numsendtime <= $endblock && $endblock > 0) ) {
                    $blocked{$id} = 'ON';
                    $numblocked ++;
                } else {
                    push(@newmsgs, { 
                        msgid    => $msgid,
                        sendtime => $sendtime,
                        shortsub => $shortsubj,
                        from     => $fromname,
                        fromdom  => $fromdom,
                        course   => $description, 
		    });
                }
            }
        }
    }
    if ($#newmsgs >= 0) {
        $r->print(<<TABLEHEAD);
<h2>$lt{'nm'}</h2>
<table class="LC_mail_list"><tr><th>&nbsp;</th>
<th>$lt{'da'}</th><th>$lt{'us'}</th><th>$lt{'do'}</th><th>$lt{'su'}</th><th>$lt{'co'}</th></tr>
TABLEHEAD
        foreach my $msg (@newmsgs) {
            $r->print(<<"ENDLINK");
<tr class="LC_mail_new">
<td><a href="/adm/email?dismode=new&amp;display=$msg->{'msgid'}">$lt{'op'}</a></td>
ENDLINK
            foreach my $item ('sendtime','from','fromdom','shortsub','course') {
                $r->print("<td>$msg->{$item}</td>");
            }
            $r->print("</tr>");
        }
        $r->print('</table>');
    } elsif ($numblocked == 0) {
        $r->print("<h3>".&mt('You have no unread messages')."</h3>");
    }
    if ($numblocked > 0) {
        my $beginblock = &Apache::lonlocal::locallocaltime($startblock);
        my $finishblock = &Apache::lonlocal::locallocaltime($endblock);
        $r->print('<h3>'.&mt('You have [quant,_1,blocked unread message,blocked unread messages].',$numblocked).'</h3>'."\n".
                 &mt('[quant,_1,message is,messages are] not viewable because display of LON-CAPA messages sent to you by other students between [_2] and [_3] is currently being blocked because of online exams.',$numblocked,$beginblock,$finishblock).'<br />'."\n".
                 &Apache::loncommon::build_block_table($startblock,$endblock,
                                                       \%setters));
    }
}


# ======================================================== Display all messages

sub disall {
    my ($r,$folder)=@_;
    $r->print(&folderlist($folder));
    if ($folder eq 'new') {
	&disnew($r);
    } elsif ($folder eq 'critical') {
	&discrit($r);
    } else {
	&disfolder($r,$folder);
    }
}

# ============================================================ Display a folder

sub disfolder {
    my ($r,$folder)=@_;
    my %blocked = ();
    my %setters = ();
    my $numblocked = 0;
    my ($startblock,$endblock) = &Apache::loncommon::blockcheck(\%setters,'com');
    $r->print(<<ENDDISHEADER);
<script type="text/javascript">
    function checkall() {
	for (i=0; i<document.forms.disall.elements.length; i++) {
            if 
          (document.forms.disall.elements[i].name.indexOf('delmark_')==0) {
	      document.forms.disall.elements[i].checked=true;
            }
        }
    }

    function uncheckall() {
	for (i=0; i<document.forms.disall.elements.length; i++) {
            if 
          (document.forms.disall.elements[i].name.indexof('delmark_')==0) {
	      document.forms.disall.elements[i].checked=false;
            }
        }
    }
</script>
ENDDISHEADER
    my $fsqs='&folder='.$folder;
    my @temp=&sortedmessages(\%blocked,$startblock,$endblock,\$numblocked,$folder);
    my $totalnumber=$#temp+1;
    unless ($totalnumber>0) {
	$r->print('<h2>'.&mt('Empty Folder').'</h2>');
	return;
    }
    unless ($interdis) {
	$interdis=20;
    }
    my $number=int($totalnumber/$interdis);
    if (($startdis<0) || ($startdis>$number)) { $startdis=$number; }
    my $firstdis=$interdis*$startdis;
    if ($firstdis>$#temp) { $firstdis=$#temp-$interdis+1; }
    my $lastdis=$firstdis+$interdis-1;
    if ($lastdis>$#temp) { $lastdis=$#temp; }
    $r->print(&scrollbuttons($startdis,$number,$firstdis,$lastdis,$totalnumber));
    $r->print('<form method="post" name="disall" action="/adm/email">'.
	      '<table class="LC_mail_list"><tr><th colspan="3">&nbsp;</th><th>');
    if ($env{'form.sortedby'} eq "revdate") {
	$r->print('<a href = "?sortedby=date'.$fsqs.'">'.&mt('Date').'</a></th>');
    } else {
	$r->print('<a href = "?sortedby=revdate'.$fsqs.'">'.&mt('Date').'</a></th>');
    }
    $r->print('<th>');
    if ($env{'form.sortedby'} eq "revuser") {
	$r->print('<a href = "?sortedby=user'.$fsqs.'">'.&mt('Username').'</a>');
    } else {
	$r->print('<a href = "?sortedby=revuser'.$fsqs.'">'.&mt('Username').'</a>');
    }
    $r->print('</th><th>');
    if ($env{'form.sortedby'} eq "revdomain") {
	$r->print('<a href = "?sortedby=domain'.$fsqs.'">'.&mt('Domain').'</a>');
    } else {
	$r->print('<a href = "?sortedby=revdomain'.$fsqs.'">'.&mt('Domain').'</a>');
    }
    $r->print('</th><th>');
    if ($env{'form.sortedby'} eq "revsubject") {
	$r->print('<a href = "?sortedby=subject'.$fsqs.'">'.&mt('Subject').'</a>');
    } else {
    	$r->print('<a href = "?sortedby=revsubject'.$fsqs.'">'.&mt('Subject').'</a>');
    }
    $r->print('</th><th>');
    if ($env{'form.sortedby'} eq "revcourse") {
        $r->print('<a href = "?sortedby=course'.$fsqs.'">'.&mt('Course').'</a>');
    } else {
        $r->print('<a href = "?sortedby=revcourse'.$fsqs.'">'.&mt('Course').'</a>');
    }
    $r->print('</th><th>');
    if ($env{'form.sortedby'} eq "revstatus") {
	$r->print('<a href = "?sortedby=status'.$fsqs.'">'.&mt('Status').'</a></th>');
    } else {
     	$r->print('<a href = "?sortedby=revstatus'.$fsqs.'">'.&mt('Status').'</a></th>');
    }
    $r->print("</tr>\n");

    my $suffix = &Apache::lonmsg::foldersuffix($folder);
    for (my $n=$firstdis;$n<=$lastdis;$n++) {
	my ($sendtime,$shortsubj,$fromname,$fromdomain,$status,$origID,
	    $description,$recv_name,$recv_domain)= 
		@{$temp[$n]};
	if (($status ne 'deleted') && defined($sendtime) && $sendtime!~/error/) {
	    if ($status eq 'new') {
		$r->print('<tr class="LC_mail_new">');
	    } elsif ($status eq 'read') {
		$r->print('<tr class="LC_mail_read">');
	    } elsif ($status eq 'replied') {
		$r->print('<tr class="LC_mail_replied">'); 
	    } else {
		$r->print('<tr class="LC_mail_other">');
	    }
	    my ($dis_name,$dis_domain) = ($fromname,$fromdomain);
	    if ($folder eq 'sent') {
		if (defined($recv_name) && !defined($recv_domain)) {
		    $dis_name   = join('<br />',@{$recv_name});
		    $dis_domain = join('<br />',@{$recv_domain});
		} else {
		    my $msg_id  = &unescape($origID);
		    my %message = &Apache::lonnet::get('nohist_email'.$suffix,
						       [$msg_id]);
		    my %content = &Apache::lonmsg::unpackagemsg($message{$msg_id});
		    $dis_name   = join('<br />',@{$content{'recuser'}});
		    $dis_domain = join('<br />',@{$content{'recdomain'}});
		}
	    }
	    $r->print('<td><input type="checkbox" name="delmark_'.$origID.'" /></td><td><a href="/adm/email?display='.$origID.$sqs. 
		      '">'.&mt('Open').'</a></td><td>'.
		      ($folder ne 'trash'?'<a href="/adm/email?markdel='.$origID.$sqs.
		      '">'.&mt('Delete'):'&nbsp').'</a></td>'.
		      '<td>'.&Apache::lonlocal::locallocaltime($sendtime).'</td><td>'.
		      $dis_name.'</td><td>'.$dis_domain.'</td><td>'.
		      $shortsubj.'</td><td>'.
                      $description.'</td><td>'.$status.'</td></tr>'."\n");
	} elsif ($status eq 'deleted') {
# purge
	    my ($result,$msg) = 
		&movemsg(&unescape($origID),$folder,'trash');
	    
	}
    }   
    $r->print("</table>\n<p>".
  '<a href="javascript:checkall()">'.&mt('Check All').'</a>&nbsp;'.
  '<a href="javascript:uncheckall()">'.&mt('Uncheck All').'</a></p>'.
  '<input type="hidden" name="sortedby" value="'.$env{'form.sortedby'}.'" />');
    if ($folder ne 'trash') {
	$r->print(
	      '<p><input type="submit" name="markeddel" value="'.&mt('Delete Checked').'" /></p>');
    }
    $r->print('<p><input type="submit" name="markedmove" value="'.&mt('Move Checked to Folder').'" />');
    my %gotfolders = &Apache::lonmsg::get_user_folders();
    my %userfolders;
    foreach my $key (keys(%gotfolders)) {
        $userfolders{$key} = $key;
    }
    $r->print(
	&Apache::loncommon::select_form('','movetofolder',
			                %userfolders));
    my $postedstartdis=$startdis+1;
    $r->print('<input type="hidden" name="folder" value="'.$folder.'" /><input type="hidden" name="startdis" value="'.$postedstartdis.'" /><input type="hidden" name="interdis" value="'.$env{'form.interdis'}.'" /></form>');
    if ($numblocked > 0) {
        my $beginblock = &Apache::lonlocal::locallocaltime($startblock);
        my $finishblock = &Apache::lonlocal::locallocaltime($endblock);
        $r->print('<br /><br />'.
                  &mt('[quant,_1,message is, messages are] not viewable because display of LON-CAPA messages sent to you by other students between [_2] and [_3] is currently being blocked because of online exams.',$numblocked,$beginblock,$finishblock));
        $r->print(&Apache::loncommon::build_block_table($startblock,$endblock,
                                                        \%setters));
    }
}

# ============================================================== Compose output

sub compout {
    my ($r,$forwarding,$replying,$broadcast,$replycrit,$folder,$dismode)=@_;
    my $suffix=&Apache::lonmsg::foldersuffix($folder);
    my ($cdom,$cnum,$group,$refarg);
    if (exists($env{'form.group'})) {
        $cdom = $env{'course.'.$env{'request.course.id'}.'.domain'};
        $cnum = $env{'course.'.$env{'request.course.id'}.'.num'};
        $group = $env{'form.group'};
        my $action = 'composing';
        $r->print(&groupmail_header($action,$group,$cdom,$cnum));
    } elsif ($broadcast eq 'individual') {
	&printheader($r,'/adm/email?compose=individual',
	     'Send a Message');
    } elsif ($broadcast) {
	&printheader($r,'/adm/email?compose=group',
	     'Broadcast Message');
    } elsif ($forwarding) {
	&Apache::lonhtmlcommon::add_breadcrumb
        ({href=>"/adm/email?display=".&escape($forwarding),
          text=>"Display Message"});
	&printheader($r,'/adm/email?forward='.&escape($forwarding),
	     'Forwarding a Message');
    } elsif ($replying) {
	&Apache::lonhtmlcommon::add_breadcrumb
        ({href=>"/adm/email?display=".&escape($replying),
          text=>"Display Message"});
	&printheader($r,'/adm/email?replyto='.&escape($replying),
	     'Replying to a Message');
    } elsif ($replycrit) {
	$r->print('<h3>'.&mt('Replying to a Critical Message').'</h3>');
	$replying=$replycrit;
    } else {
	&printheader($r,'/adm/email?compose=upload',
	     'Distribute from Uploaded File');
    }

    my $dispcrit='';
    my $dissub='';
    my $dismsg='';
    my $disbase='';
    my $func=&mt('Send New');
    my %lt=&Apache::lonlocal::texthash('us'  => 'Username',
				       'do'  => 'Domain',
				       'ad'  => 'Additional Recipients',
				       'sb'  => 'Subject',
				       'ca'  => 'Cancel',
				       'ma'  => 'Mail',
                                       'gen' => 'Generate messages from a file',
                                       'gmt' => 'General message text',
                                       'tff' => 'The file format for the uploaded portion of the message is',
                                       'uas' => 'Upload and Send',
                                      );
    if (&Apache::lonnet::allowed('srm',$env{'request.course.id'})
	|| &Apache::lonnet::allowed('srm',$env{'request.course.id'}.
				    '/'.$env{'request.course.sec'})) {
	 my $crithelp = Apache::loncommon::help_open_topic("Course_Critical_Message");
         $dispcrit=
 '<p><label><input type="checkbox" name="critmsg" /> '.&mt('Send as critical message').'</label> ' . $crithelp . 
 '</p><p>'.
 '<label><input type="checkbox" name="sendbck" /> '.&mt('Send as critical message').'  ' .
 &mt('and return receipt') . '</label>' . $crithelp . 
 '</p><p><label><input type="checkbox" name="permanent" /> '.
&mt('Send copy to permanent email address (if known)').'</label></p>'.
'<p><label><input type="checkbox" name="rsspost" /> '.
		  &mt('Include in course RSS newsfeed').'</label></p>';
     }
    my %message;
    my %content;
    my $defdom=$env{'user.domain'};
    if ($forwarding) {
	%message=&Apache::lonnet::get('nohist_email'.$suffix,[$forwarding]);
	%content=&Apache::lonmsg::unpackagemsg($message{$forwarding},$folder);
	$dispcrit.='<input type="hidden" name="forwid" value="'.
	    $forwarding.'" />';
	$func=&mt('Forward');
	
	$dissub=&mt('Forwarding').': '.$content{'subject'};
	$dismsg=&mt('Forwarded message from').' '.
	    $content{'sendername'}.' '.&mt('at').' '.$content{'senderdomain'};
	if ($content{'baseurl'}) {
	    $disbase='<input type="hidden" name="baseurl" value="'.&escape($content{'baseurl'}).'" />';
	}
    }
    if ($replying) {
	%message=&Apache::lonnet::get('nohist_email'.$suffix,[$replying]);
	%content=&Apache::lonmsg::unpackagemsg($message{$replying},$folder);
	$dispcrit.='<input type="hidden" name="replyid" value="'.
	    $replying.'" />';
	$func=&mt('Send Reply to');
	
	$dissub=&mt('Reply').': '.$content{'subject'};       
	$dismsg='> '.$content{'message'};
	$dismsg=~s/\r/\n/g;
	$dismsg=~s/\f/\n/g;
	$dismsg=~s/\n+/\n\> /g;
	if ($content{'baseurl'}) {
	    $disbase='<input type="hidden" name="baseurl" value="'.&escape($content{'baseurl'}).'" />';
	    if ($env{'user.adv'}) {
		$disbase.='<label><input type="checkbox" name="storebasecomment" />'.&mt('Store message for re-use').
		    '</label> <a href="/adm/email?showcommentbaseurl='.
		    &escape($content{'baseurl'}).'" target="comments">'.
		    &mt('Show re-usable messages').'</a><br />';
	    }
	}
    }
    my $citation=&displayresource(%content);
    my ($can_grp_broadcast,$viewgrps,$editgrps);
    if ($env{'form.recdom'}) { $defdom=$env{'form.recdom'}; }
    if ($env{'form.text'}) { $dismsg=$env{'form.text'}; }
    if ($env{'form.subject'}) { $dissub=$env{'form.subject'}; }
    $r->print(
                '<form action="/adm/email"  name="compemail" method="post"'.
                ' enctype="multipart/form-data">'."\n".
                '<input type="hidden" name="sendmail" value="on" />'."\n");
    if ($broadcast eq 'group' && $env{'form.group'} ne '') {
        $can_grp_broadcast = 
                &Apache::lonnet::allowed('sgb',$env{'request.course.id'}.'/'.
                                         $group);
        $viewgrps = 
               &Apache::lonnet::allowed('vcg',$env{'request.course.id'}.
               ($env{'request.course.sec'}?'/'.$env{'request.course.sec'}:''));
        $editgrps = 
               &Apache::lonnet::allowed('mdg',$env{'request.course.id'}.
               ($env{'request.course.sec'}?'/'.$env{'request.course.sec'}:''));
        if ($viewgrps || $editgrps || $can_grp_broadcast) {
            $r->print(&disgroup($cdom,$cnum,$group,$viewgrps,$editgrps));
        }
    }
    $r->print('<table>');
    if (($broadcast eq 'group') && ($group ne '') && 
        (!$can_grp_broadcast && !$viewgrps && !$editgrps)) {
        $r->print(&recipient_input_row($cdom,%lt));
    } 
    if (($broadcast ne 'group') && ($broadcast ne 'upload')) {
	if ($replying) {
	    $r->print('<tr><td colspan="2">'.&mt('Replying to').' '.
		      &Apache::loncommon::aboutmewrapper(
							 &Apache::loncommon::plainname($content{'sendername'},$content{'senderdomain'}),$content{'sendername'},$content{'senderdomain'}).' ('.
		      $content{'sendername'}.':'.
		      $content{'senderdomain'}.')'.
		      '<input type="hidden" name="recuname" value="'.$content{'sendername'}.'" />'.
		      '<input type="hidden" name="recdomain" value="'.$content{'senderdomain'}.'" />'.
		      '</td></tr>');
	} else {
            $r->print(&recipient_input_row($defdom,%lt));
        }
    }
    my $latexHelp = Apache::loncommon::helpLatexCheatsheet();
    if ($broadcast ne 'upload') {
       $r->print(<<"ENDCOMP");
<tr><td>$lt{'ad'}:<br /><tt>username:domain,username:domain, ...
</tt></td><td>
<input type="text" size="50" name="additionalrec" /></td></tr>
<tr><td>$lt{'sb'}:</td><td><input type="text" size="50" name="subject" value="$dissub" />
</td></tr></table>
$latexHelp
<textarea name="message" id="message" cols="80" rows="15" wrap="hard">$dismsg
</textarea></p><br />
$dispcrit
$disbase
<input type="hidden" name="folder" value="$folder" />
<input type="hidden" name="dismode" value="$dismode" />
<input type="submit" name="send" value="$func $lt{'ma'}" />
<input type="submit" name="cancel" value="$lt{'ca'}" /><hr />
$citation
ENDCOMP
        if (exists($env{'form.ref'})) {
            $r->print('<input type="hidden" name="ref" value="'.
                      $env{'form.ref'}.'" />');
        }
        if (exists($env{'form.group'})) {
            $r->print('<input type="hidden" name="group" value="'.
                      $env{'form.group'}.'" />');
        }
    } else { # $broadcast is 'upload'
	$r->print(<<ENDBLOCK);
<input type="hidden" name="sendmode" value="upload" />
<input type="hidden" name="send" value="on" />
<h3>$lt{'gen'}</h3>
<p>
Subject: <input type="text" size="50" name="subject" />
</p>
<p>$lt{'gmt'}:<br />
<textarea name="message" id="message" cols="60" rows="10" wrap="hard">$dismsg
</textarea></p>
<p>
$lt{'tff'}:
ENDBLOCK
       $r->print('
<pre>'."\n".
&mt('username1:domain1: text')."\n".
&mt('username2:domain2: text')."\n".
&mt('username3:domain1: text')."\n".
'</pre>
</p>
<p>
'.&mt('The messages will be assembled from all lines with the respective'."\n".'<tt>username:domain</tt>, and appended to the general message text.'));
        $r->print(<<ENDUPLOAD);
</p>
<p>
<input type="file" name="upfile" size="40" /></p><p>
$dispcrit
<input type="submit" value="$lt{'uas'}" /></p>
ENDUPLOAD
    }
    if ($broadcast eq 'group') {
       if ($group eq '') {
           my $studentsel = &discourse();
           $r->print($studentsel);
       }
    }
    if ($env{'form.displayedcrit'}) {
	$r->print('<input type="hidden" name="displayedcrit" value="true" />');
    }
    $r->print('</form>'.
	      &Apache::lonfeedback::generate_preview_button('compemail','message').
	      &Apache::lonhtmlcommon::htmlareaselectactive('message'));
}

# ---------------------------------------------------- Display all face to face

sub recipient_input_row {
    my ($dom,%lt) = @_;
    my $domform = &Apache::loncommon::select_dom_form($dom,'recdomain');
    my $selectlink=
      &Apache::loncommon::selectstudent_link('compemail','recuname',
                                             'recdomain');
    my $output = <<"ENDREC";
<tr><td>$lt{'us'}:</td><td><input type="text" size="12" name="recuname" value="$env{'form.recname'}" /></td><td rowspan="2">$selectlink</td></tr>
<tr><td>$lt{'do'}:</td>
<td>$domform</td></tr>
ENDREC
    return $output;
}

sub retrieve_instructor_comments {
    my ($user,$domain)=@_;
    my $target=$env{'form.grade_target'};
    if (! $env{'request.course.id'}) { return; }
    if (! &Apache::lonnet::allowed('dff',$env{'request.course.id'})
	&& ! &Apache::lonnet::allowed('dff',$env{'request.course.id'}.
				      '/'.$env{'request.course.sec'})) {
	return;
    }
    my %records=&Apache::lonnet::dump('nohist_email',
			 $env{'course.'.$env{'request.course.id'}.'.domain'},
			 $env{'course.'.$env{'request.course.id'}.'.num'},
                         '%255b'.$user.'%253a'.$domain.'%255d');
    my $result='';
    foreach my $key (sort(keys(%records))) {
        my %content=&Apache::lonmsg::unpackagemsg($records{$key});
        next if ($content{'senderdomain'} eq '');
        next if ($content{'subject'} !~ /^Record/);
	# &Apache::lonfeedback::newline_to_br(\$content{'message'});
	$result.='Recorded by '.
            $content{'sendername'}.':'.$content{'senderdomain'}."\n";
        $result.=
            &Apache::lontexconvert::msgtexconverted($content{'message'})."\n";
     }
    return $result;
}

sub disfacetoface {
    my ($r,$user,$domain)=@_;
    my $target=$env{'form.grade_target'};
    unless ($env{'request.course.id'}) { return; }
    if  (!&Apache::lonnet::allowed('dff',$env{'request.course.id'})
	 && ! &Apache::lonnet::allowed('dff',$env{'request.course.id'}.
				       '/'.$env{'request.course.sec'})) {
	$r->print(&mt('Not allowed'));
	return;
    }
    my %records=&Apache::lonnet::dump('nohist_email',
			 $env{'course.'.$env{'request.course.id'}.'.domain'},
			 $env{'course.'.$env{'request.course.id'}.'.num'},
                         '%255b'.$user.'%253a'.$domain.'%255d');
    my $result='';
    foreach my $key (sort(keys(%records))) {
        my %content=&Apache::lonmsg::unpackagemsg($records{$key});
        next if ($content{'senderdomain'} eq '');
	&Apache::lonfeedback::newline_to_br(\$content{'message'});
        if ($content{'subject'}=~/^Record/) {
	    $result.='<h3>'.&mt('Record').'</h3>';
        } elsif ($content{'subject'}=~/^Broadcast/) {
            $result .='<h3>'.&mt('Broadcast Message').'</h3>';
            if ($content{'subject'}=~/^Broadcast\./) {
                if (defined($content{'coursemsgid'})) {
                    my $crsmsgid = &escape($content{'coursemsgid'});
                    my $broadcast_message = &general_message($crsmsgid);
                    $content{'message'} = '<b>'.&mt('Subject').': '.$content{'message'}.'</b><br />'.$broadcast_message;
                } else {
                    %content=&Apache::lonmsg::unpackagemsg($content{'message'});
                    $content{'message'} =
                    '<b>'.&mt('Subject').': '.$content{'subject'}.'</b><br />'.
                    $content{'message'};
                }
            }    
        } else {
            $result.='<h3>'.&mt('Critical Message').'</h3>';
            if (defined($content{'coursemsgid'})) {
                my $crsmsgid=&escape($content{'coursemsgid'});
                my $critical_message = &general_message($crsmsgid);
                $content{'message'} = '<b>'.&mt('Subject').': '.$content{'message'}.'</b><br />'.$critical_message;
            } else {
                %content=&Apache::lonmsg::unpackagemsg($content{'message'});
                $content{'message'}=
                '<b>'.&mt('Subject').': '.$content{'subject'}.'</b><br />'.
		$content{'message'};
            }
        }
        $result.=&mt('By').': <b>'.
&Apache::loncommon::aboutmewrapper(
 &Apache::loncommon::plainname($content{'sendername'},$content{'senderdomain'}),$content{'sendername'},$content{'senderdomain'}).'</b> ('.
$content{'sendername'}.':'.
            $content{'senderdomain'}.') '.$content{'time'}.
            '<br /><pre>'.
              &Apache::lontexconvert::msgtexconverted($content{'message'}).
	      '</pre>';
     }
    # Check to see if there were any messages.
    if ($result eq '') {
        my $lctype = lc(&Apache::loncommon::course_type());
	if ($target ne 'tex') { 
	    $r->print("<p><b>".&mt('No notes, face-to-face discussion records, critical messages, or broadcast messages in this [_1].',$lctype)."</b></p>");
	} else {
	    $r->print('\textbf{'.&mt('No notes, face-to-face discussion records, critical messages or broadcast messages in this [_1].',$lctype).'}\\\\');
	}
    } else {
       $r->print($result);
    }
}

sub general_message {
    my ($crsmsgid) = @_;
    my %general_content;
    if ($crsmsgid) { 
        my %course_content = &Apache::lonnet::get('nohist_email',[$crsmsgid],
                           $env{'course.'.$env{'request.course.id'}.'.domain'},
                           $env{'course.'.$env{'request.course.id'}.'.num'});
        %general_content = &Apache::lonmsg::unpackagemsg($course_content{$crsmsgid});
    }
    return $general_content{'message'};
}

# ---------------------------------------------------------------- Face to face

sub facetoface {
    my ($r,$stage)=@_;
    if (!&Apache::lonnet::allowed('dff',$env{'request.course.id'})
	&& ! &Apache::lonnet::allowed('dff',$env{'request.course.id'}.
				      '/'.$env{'request.course.sec'})) {
	$r->print(&mt('Not allowed'));
	return;
    }
    my $crstype = &Apache::loncommon::course_type();
    my $leaders = ($crstype eq 'Group') ? 'coordinators and leaders'
                                        : 'faculty and staff';
    &printheader($r,
		 '/adm/email?recordftf=query',
		 "User Notes, Face-to-Face, Critical Messages, Broadcast Messages");
# from query string

    if ($env{'form.recname'}) { $env{'form.recuname'}=$env{'form.recname'}; }
    if ($env{'form.recdom'}) { $env{'form.recdomain'}=$env{'form.recdom'}; }

    my $defdom=$env{'user.domain'};
# already filled in
    if ($env{'form.recdomain'}) { $defdom=$env{'form.recdomain'}; }
# generate output
    my $domform = &Apache::loncommon::select_dom_form($defdom,'recdomain');
    my $stdbrws = &Apache::loncommon::selectstudent_link
	('stdselect','recuname','recdomain');
    my %lt=&Apache::lonlocal::texthash('user' => 'Username',
				       'dom' => 'Domain',
				       'head' => "User Notes, Records of Face-To-Face Discussions, Critical Messages, and Broadcast Messages in $crstype",
				       'subm' => 'Retrieve discussion and message records',
				       'newr' => 'New Record (record is visible to '.lc($crstype).' '.$leaders.')',
				       'post' => 'Post this Record');
    $r->print(<<"ENDTREC");
<h3>$lt{'head'}</h3>
<form method="post" action="/adm/email" name="stdselect">
<input type="hidden" name="recordftf" value="retrieve" />
<table>
<tr><td>$lt{'user'}:</td><td><input type="text" size="12" name="recuname" value="$env{'form.recuname'}" /></td>
<td rowspan="2">
$stdbrws
<input type="submit" value="$lt{'subm'}" /></td>
</tr>
<tr><td>$lt{'dom'}:</td>
<td>$domform</td></tr>
</table>
</form>
ENDTREC
    if (($stage ne 'query') &&
        ($env{'form.recdomain'}) && ($env{'form.recuname'})) {
        chomp($env{'form.newrecord'});
        if ($env{'form.newrecord'}) {
	    &Apache::lonmsg::store_instructor_comment($env{'form.newrecord'},
						      $env{'form.recuname'},
						      $env{'form.recdomain'});
        }
        $r->print('<h3>'.&Apache::loncommon::plainname($env{'form.recuname'},
				     $env{'form.recdomain'}).'</h3>');
        &disfacetoface($r,$env{'form.recuname'},$env{'form.recdomain'});
	$r->print(<<ENDRHEAD);
<form method="post" action="/adm/email">
<input name="recdomain" value="$env{'form.recdomain'}" type="hidden" />
<input name="recuname" value="$env{'form.recuname'}" type="hidden" />
ENDRHEAD
        $r->print(<<ENDBFORM);
<hr />$lt{'newr'}<br />
<textarea name="newrecord" cols="80" rows="10" wrap="hard"></textarea>
<br />
<input type="hidden" name="recordftf" value="post" />
<input type="submit" value="$lt{'post'}" />
</form>
ENDBFORM
    }
}

# ----------------------------------------------------------- Blocking during exams

sub examblock {
    my ($r,$action) = @_;
    unless ($env{'request.course.id'}) { return;}
    if (!&Apache::lonnet::allowed('dcm',$env{'request.course.id'})
	&& ! &Apache::lonnet::allowed('dcm',$env{'request.course.id'}.
				      '/'.$env{'request.course.sec'})) {
	$r->print('Not allowed');
	return;
    }
    my $usertype = (&Apache::loncommon::course_type() eq 'Group') ? 'members'
	                                                          : 'students';
    my %lt=&Apache::lonlocal::texthash(
            'comb' => 'Communication Blocking',
            'cbds' => 'Communication blocking during scheduled exams',
            'desc' => "You can use communication blocking to prevent $usertype enrolled in this course from displaying LON-CAPA messages sent by other $usertype during an online exam. As blocking of communication could potentially interrupt legitimate communication between $usertype who are also both enrolled in a different LON-CAPA course, please be careful that you select the correct start and end times for your scheduled exam when setting or modifying these parameters.",
             'mecb' => 'Modify existing communication blocking periods',
             'ncbc' => 'No communication blocks currently stored',
             'stor' => 'Store',
    );

    my %ltext = &Apache::lonlocal::texthash(
            'dura' => 'Duration',
            'setb' => 'Set by',
            'even' => 'Event',
            'blck' => 'Blocked?',
            'actn' => 'Action',
            'star' => 'Start',
            'endd' => 'End'
    );

    &printheader($r,'/adm/email?block=display',$lt{'comb'});
    $r->print('<h3>'.$lt{'cbds'}.'</h3>');

    if ($action eq 'store') {
        &blockstore($r);
    }

    $r->print($lt{'desc'}.'<br /><br />
               <form name="blockform" method="post" action="/adm/email?block=store">
             ');

    $r->print('<h4>'.$lt{'mecb'}.'</h4>');
    my %records = ();
    my $blockcount = 0;
    my $parmcount = 0;
    &get_blockdates(\%records,\$blockcount);
    if ($blockcount > 0) {
        $parmcount = &display_blocker_status($r,\%records,\%ltext);
    } else {
        $r->print($lt{'ncbc'}.'<br /><br />');
    }
    &display_addblocker_table($r,$parmcount,\%ltext);
    my $end_page=&Apache::loncommon::end_page();
    $r->print(<<"END");
<br />
<input type="hidden" name="blocktotal" value="$blockcount" />
<input type ="submit" value="$lt{'stor'}" />
</form>
$end_page
END
    return;
}

sub blockstore {
    my $r = shift;
    my %lt=&Apache::lonlocal::texthash(
            'tfcm' => 'The following changes were made',
            'ncwm' => 'No changes were made.' 
    );
    my %adds = ();
    my %removals = ();
    my %cancels = ();
    my $modtotal = 0;
    my $canceltotal = 0;
    my $addtotal = 0;
    my %blocking = ();
    $r->print('<h3>'.$lt{'head'}.'</h3>');
    foreach my $envkey (keys(%env)) {
        if ($envkey =~ m/^form\.modify_(\d+)$/) {
            $adds{$1} = $1;
            $removals{$1} = $1;
            $modtotal ++;
        } elsif ($envkey =~ m/^form\.cancel_(\d+)$/) {
            $cancels{$1} = $1;
            unless ( defined($removals{$1}) ) {
                $removals{$1} = $1;
                $canceltotal ++;
            }
        } elsif ($envkey =~ m/^form\.add_(\d+)$/) {
            $adds{$1} = $1;
            $addtotal ++;
        } 
    }

    foreach my $key (keys(%removals)) {
        my $hashkey = $env{'form.key_'.$key};
        &Apache::lonnet::del('comm_block',["$hashkey"],
                         $env{'course.'.$env{'request.course.id'}.'.domain'},
                         $env{'course.'.$env{'request.course.id'}.'.num'}
                         );
    }
    foreach my $key (keys(%adds)) {
        unless ( defined($cancels{$key}) ) {
            my ($newstart,$newend) = &get_dates_from_form($key);
            my $newkey = $newstart.'____'.$newend;
            my $blocktypes = &get_block_choices($key);
            $blocking{$newkey} = {
                          setter => $env{'user.name'}.':'.$env{'user.domain'},
                          event  => &escape($env{'form.title_'.$key}),
                          blocks => $blocktypes,
                        };
        }
    }
    if ($addtotal + $modtotal > 0) {
        &Apache::lonnet::put('comm_block',\%blocking,
                     $env{'course.'.$env{'request.course.id'}.'.domain'},
                     $env{'course.'.$env{'request.course.id'}.'.num'}
                     );
    }
    my $chgestotal = $canceltotal + $modtotal + $addtotal;
    if ($chgestotal > 0) {
        $r->print($lt{'tfcm'}.'<ul>');
        if ($canceltotal > 0) {
            $r->print('<li>'.&mt('[quant,_1,communication blocking period was,communication blocking periods were] removed.',$canceltotal).'</li>');
        }
        if ($modtotal > 0) {
            $r->print('<li>'.&mt('[quant,_1,communication blocking period was,communication blocking periods were] modified.',$modtotal).'</li>');
        }
        if ($addtotal > 0) {
            $r->print('<li>'.&mt('[quant,_1,communication blocking period was,communication blocking periods were] added.',$addtotal).'</li>');
        }
        $r->print('</ul>');
    } else {
        $r->print($lt{'ncwm'});
    }
    $r->print('<br />');
    return;
}

sub get_dates_from_form {
    my $item = shift;
    my $startdate = &Apache::lonhtmlcommon::get_date_from_form('startdate_'.$item);
    my $enddate   = &Apache::lonhtmlcommon::get_date_from_form('enddate_'.$item);
    return ($startdate,$enddate);
}

sub get_blockdates {
    my ($records,$blockcount) = @_;
    $$blockcount = 0;
    %{$records} = &Apache::lonnet::dump('comm_block',
                         $env{'course.'.$env{'request.course.id'}.'.domain'},
                         $env{'course.'.$env{'request.course.id'}.'.num'}
                         );
    $$blockcount = keys(%{$records});

    if ((keys(%{$records}))[0] =~ /^error: 2 /) {
	$records = {};
	$$blockcount = 0;
    }
}

sub get_block_choices {
    my $item = shift;
    my $blocklist;
    my ($typeorder,$types) = &blocktype_text();
    foreach my $type (@{$typeorder}) {
        if ($env{'form.'.$type.'_'.$item}) {
            $blocklist->{$type} = 'on'; 
        } else {
            $blocklist->{$type} = 'off';
        }
    }
    return $blocklist;
}

sub display_blocker_status {
    my ($r,$records,$ltext) = @_;
    my $parmcount = 0;
  
    my %lt = &Apache::lonlocal::texthash(
        'modi' => 'Modify',
        'canc' => 'Cancel',
    );
    my ($typeorder,$types) = &blocktype_text();
    $r->print(&Apache::loncommon::start_data_table());
    $r->print(<<"END");
  <tr>
    <th>$ltext->{'dura'}</th>
    <th>$ltext->{'setb'}</th>
    <th>$ltext->{'even'}</th>
    <th>$ltext->{'blck'}</th>
    <th>$ltext->{'actn'}?</th>
  </tr>
END
    foreach my $record (sort(keys(%{$records}))) {
        my $onchange = 'onFocus="javascript:window.document.forms['.
                       "'blockform'].elements['modify_".$parmcount."'].".
                       'checked=true;"';
        my ($start,$end) = split(/____/,$record);
        my $startform = &Apache::lonhtmlcommon::date_setter('blockform','startdate_'.$parmcount,$start,$onchange);
        my $endform = &Apache::lonhtmlcommon::date_setter('blockform','enddate_'.$parmcount,$end,$onchange);
	
	my ($setuname,$setudom,$title,$blocks) = 
	    &Apache::loncommon::parse_block_record($$records{$record});
	$title = &HTML::Entities::encode($title,'"<>&');
        my $settername = 
           &Apache::loncommon::aboutmewrapper(
                           &Apache::loncommon::plainname($setuname,$setudom),
                           $setuname,$setudom);
        $r->print(&Apache::loncommon::start_data_table_row());
        $r->print(<<"END");
        <td>$ltext->{'star'}:&nbsp;$startform<br/>$ltext->{'endd'}:&nbsp;&nbsp;$endform</td>
        <td>$settername</td>
        <td><input type="text" name="title_$parmcount" size="15" value="$title" /><input type="hidden" name="key_$parmcount" value="$record" /></td>
        <td>
END
        foreach my $block (@{$typeorder}) {
            my $blockstatus = '';
            if ($blocks->{$block} eq 'on') {
                $blockstatus = 'checked="true"';
            }
            $r->print('<label><input type="checkbox" name="'.$block.'_'.$parmcount.'" '.$blockstatus.' value="1" />'.$types->{$block}.'</label><br />');
        }
        $r->print(<<"END");
        </td>      
        <td><label>$lt{'modi'}?&nbsp;<input type="checkbox" name="modify_$parmcount" /></label><br /><label>$lt{'canc'}?&nbsp;&nbsp;<input type="checkbox" name="cancel_$parmcount" /></label>
END
        $r->print(&Apache::loncommon::end_data_table_row());
        $parmcount++;
    }
    $r->print(<<"END");
</table>
<br />
<br />
END
    return $parmcount;
}

sub display_addblocker_table {
    my ($r,$parmcount,$ltext) = @_;
    my $start = time;
    my $end = $start + (60 * 60 * 2); #Default is an exam of 2 hours duration.
    my $onchange = 'onFocus="javascript:window.document.forms['.
                   "'blockform'].elements['add_".$parmcount."'].".
                   'checked=true;"';
    my $startform = &Apache::lonhtmlcommon::date_setter('blockform','startdate_'.$parmcount,$start,$onchange);
    my $endform = &Apache::lonhtmlcommon::date_setter('blockform','enddate_'.$parmcount,$end,$onchange);
    my %lt = &Apache::lonlocal::texthash(
        'addb' => 'Add block',
        'exam' => 'e.g., Exam 1',
        'addn' => 'Add new communication blocking periods'
    );
    my ($typeorder,$types) = &blocktype_text();
    $r->print(<<"END");
<h4>$lt{'addn'}</h4> 
END
    $r->print(&Apache::loncommon::start_data_table());
    $r->print(<<"END");
   <tr>
     <th>$ltext->{'dura'}</th>
     <th>$ltext->{'even'} $lt{'exam'}</th>
     <th>$ltext->{'blck'}</th>
     <th>$ltext->{'actn'}?</th>
   </tr>
END
    $r->print(&Apache::loncommon::start_data_table_row());
    $r->print(<<"END");
     <td>$ltext->{'star'}:&nbsp;$startform<br />$ltext->{'endd'}:&nbsp;&nbsp;$endform</td>
     <td><input type="text" name="title_$parmcount" size="15" value="" /></td>
     <td>
END
    foreach my $block (@{$typeorder}) {
        $r->print('<label><input type="checkbox" name="'.$block.'_'.$parmcount.'" value="1" />'.$types->{$block}.'</label><br />');
     }
     $r->print(<<"END");
     </td> 
     <td><label>$lt{'addb'}?&nbsp;<input type="checkbox" name="add_$parmcount" value="1" /></label></td>
END
    $r->print(&Apache::loncommon::end_data_table_row());
    $r->print(&Apache::loncommon::end_data_table());
    return;
}

sub blocktype_text {
    my %types = &Apache::lonlocal::texthash(
        'com' => 'Messaging',
        'chat' => 'Chat',
        'boards' => 'Discussion',
        'port' => 'Portfolio',
        'groups' => 'Groups',
        'blogs' => 'Blogs',
    );
    my $typeorder = ['com','chat','boards','port','groups','blogs'];
    return ($typeorder,\%types);
}

# ----------------------------------------------------------- Display a message

sub displaymessage {
    my ($r,$msgid,$folder)=@_;
    my $suffix=&Apache::lonmsg::foldersuffix($folder);
    my %blocked = ();
    my %setters = ();
    my $numblocked = 0;
    my $crstype = &Apache::loncommon::course_type();

# info to generate "next" and "previous" buttons and check if message is blocked
    my ($startblock,$endblock) = &Apache::loncommon::blockcheck(\%setters,'com');
    my @messages=&sortedmessages(\%blocked,$startblock,$endblock,\$numblocked,$folder);
    if ( $blocked{$msgid} eq 'ON' ) {
        &printheader($r,'/adm/email',&mt('Display a Message'));
        $r->print(&mt('You attempted to display a message that is currently blocked because you are enrolled in one or more courses for which there is an ongoing online exam.'));
        &build_block_table($r,$startblock,$endblock,\%setters);
        return;
    }
    &statuschange($msgid,'read',$folder);
    my %message=&Apache::lonnet::get('nohist_email'.$suffix,[$msgid]);
    my %content=&Apache::lonmsg::unpackagemsg($message{$msgid});

    my $counter=0;
    $r->print('<pre>');
    my $escmsgid=&escape($msgid);
    foreach (@messages) {
	if ($_->[5] eq $escmsgid){
	    last;
	}
	$counter++;
    }
    $r->print('</pre>');
    my $number_of_messages = scalar(@messages); #subtract 1 for last index
# start output
    &printheader($r,'/adm/email?display='.&escape($msgid),'Display a Message','',$content{'baseurl'});
    my %courseinfo=&Apache::lonnet::coursedescription($content{'courseid'});
# Functions
    $r->print('<table border="2" width="100%"><tr bgcolor="#FFFFAA"><td>'.&mt('Functions').':</td>'.
	      '<td><a href="/adm/email?replyto='.&escape($msgid).$sqs.
	      '"><b>'.&mt('Reply').'</b></a></td>'.
	      '<td><a href="/adm/email?forward='.&escape($msgid).$sqs.
	      '"><b>'.&mt('Forward').'</b></a></td>'.
	      '<td><a href="/adm/email?markunread='.&escape($msgid).$sqs.
	      '"><b>'.&mt('Mark Unread').'</b></a></td>'.
	      '<td><a href="/adm/email?markdel='.&escape($msgid).$sqs.
	      '"><b>'.&mt('Delete').'</b></a></td>'.
	      '<td><a href="/adm/email?'.$sqs.
	      ($env{'form.dismode'} eq 'new'?'&folder=new':'').
	      '"><b>'.&mt('Back to Folder Display').'</b></a></td>');
    if ($counter > 0){
	$r->print('<td><a href="/adm/email?display='.$messages[$counter-1]->[5].$sqs.
		  '"><b>'.&mt('Previous').'</b></a></td>');
    }
    if ($counter < $number_of_messages - 1){
	$r->print('<td><a href="/adm/email?display='.$messages[$counter+1]->[5].$sqs.
		  '"><b>'.&mt('Next').'</b></a></td>');
    }
    $r->print('</tr></table>');
    if ($env{'user.adv'}) {
	$r->print('<table border="2" width="100%"><tr bgcolor="#FFAAAA"><td>'.&mt('Currently available actions (will open extra window)').':</td>');
	my $symb=&Apache::lonnet::symbread($content{'baseurl'});      
	if (&Apache::lonnet::allowed('vgr',$env{'request.course.id'})) {
		$r->print('<td><b>'.&Apache::loncommon::track_student_link(&mt('View recent activity'),$content{'sendername'},$content{'senderdomain'},'check').'</b></td>');
	    }
	if (&Apache::lonnet::allowed('opa',$env{'request.course.id'}) && $symb) {
	    $r->print('<td><b>'.&Apache::loncommon::pprmlink(&mt('Set/Change parameters'),$content{'sendername'},$content{'senderdomain'},$symb,'check').'</b></td>');
	}
	if (&Apache::lonnet::allowed('mgr',$env{'request.course.id'}) && $symb) {
	    $r->print('<td><b>'.&Apache::loncommon::pgrdlink(&mt('Set/Change grades'),$content{'sendername'},$content{'senderdomain'},$symb,'check').'</b></td>');
	}
	$r->print('</tr></table>');
    }
    my $tolist;
    my @recipients = ();
    for (my $i=0; $i<@{$content{'recuser'}}; $i++) {
        $recipients[$i] =  &Apache::loncommon::aboutmewrapper(
           &Apache::loncommon::plainname($content{'recuser'}[$i],
                                      $content{'recdomain'}[$i]),
              $content{'recuser'}[$i],$content{'recdomain'}[$i]).
       ' ('.$content{'recuser'}[$i].' at '.$content{'recdomain'}[$i].') ';
    }
    $tolist = join(', ',@recipients);
    $r->print('<br /><b>'.&mt('Subject').':</b> '.$content{'subject'}.
	      ($folder ne 'sent'?'<br /><b>'.&mt('From').':</b> '.
	      &Apache::loncommon::aboutmewrapper(
						 &Apache::loncommon::plainname($content{'sendername'},$content{'senderdomain'}),
						 $content{'sendername'},$content{'senderdomain'}).' ('.
	      $content{'sendername'}.' at '.
	      $content{'senderdomain'}.') ':'<br /><b>'.&mt('To').':</b> '.
              $tolist).
	      ($content{'courseid'}?'<br /><b>'.&mt($crstype).':</b> '.$courseinfo{'description'}.
	       ($content{'coursesec'}?' ('.&mt('Section').': '.$content{'coursesec'}.')':''):'').
	      '<br /><b>'.&mt('Time').':</b> '.$content{'time'}.
	      ($content{'baseurl'}?'<br /><b>'.&mt('Refers to').':</b> <a href="'.$content{'baseurl'}.'">'.
	       $content{'baseurl'}.' ('.&Apache::lonnet::gettitle($content{'baseurl'}).')</a>':'').
	      '<p><pre>'.
	      &Apache::lontexconvert::msgtexconverted($content{'message'},1).
	      '</pre><hr />'.&displayresource(%content).'</p>');
    return;   
}

# =========================================================== Show the citation

sub displayresource {
    my %content=@_;
#
# If the recipient is in the same course that the message was sent from and
# has sufficient privileges, show "all details," else show citation
#
    if (($env{'request.course.id'} eq $content{'courseid'})
     && (&Apache::lonnet::allowed('vgr',$content{'courseid'}))) {
	my $symb=&Apache::lonnet::symbread($content{'baseurl'});
# Could not get a symb, give up
	unless ($symb) { return $content{'citation'}; }
# Have a symb, can render
	return '<h2>'.&mt('Current attempts of student (if applicable)').'</h2>'.
	    &Apache::loncommon::get_previous_attempt($symb,
						     $content{'sendername'},
						     $content{'senderdomain'},
						     $content{'courseid'}).
	    '<hr /><h2>'.&mt('Current screen output (if applicable)').'</h2>'.
	    &Apache::loncommon::get_student_view($symb,
						 $content{'sendername'},
						 $content{'senderdomain'},
						 $content{'courseid'}).
	    '<h2>'.&mt('Correct Answer(s) (if applicable)').'</h2>'.
	    &Apache::loncommon::get_student_answers($symb,
						    $content{'sendername'},
						    $content{'senderdomain'},
						    $content{'courseid'});
    } elsif ($env{'user.adv'}) {
	return $content{'citation'};
    }
    return '';
}

# ================================================================== The Header

sub header {
    my ($r,$title,$baseurl)=@_;
    
    my $extra = &Apache::loncommon::studentbrowser_javascript();
    if ($baseurl) {
	$extra .= "<base href=\"".&Apache::lonnet::absolute_url()."/$baseurl\" />";
    }
    $r->print(&Apache::loncommon::start_page('Communication and Messages',
 					$extra));
    $r->print(&Apache::lonhtmlcommon::breadcrumbs
     		(($title?$title:'Communication and Messages')));
}

# ---------------------------------------------------------------- Print header

sub printheader {
    my ($r,$url,$desc,$title,$baseurl)=@_;
    &Apache::lonhtmlcommon::add_breadcrumb
	({href=>$url,
	  text=>$desc});
    &header($r,$title,$baseurl);
}

# ------------------------------------------------------------ Store the comment

sub storecomment {
    my ($r)=@_;
    my $msgtxt=&Apache::lonfeedback::clear_out_html($env{'form.message'});
    my $cleanmsgtxt='';
    foreach my $line (split(/[\n\r]/,$msgtxt)) {
	unless ($line=~/^\s*(\>|\&gt\;)/) {
	    $cleanmsgtxt.=$line."\n";
	}
    }
    my $key=&escape($env{'form.baseurl'}).'___'.time;
    &Apache::lonnet::put('nohist_stored_comments',{ $key => $cleanmsgtxt });
}

sub storedcommentlisting {
    my ($r)=@_;
    my %msgs=&Apache::lonnet::dump('nohist_stored_comments',undef,undef,
       '^'.&escape(&escape($env{'form.showcommentbaseurl'})));
    $r->print(&Apache::loncommon::start_page('Stored Comment Listing',undef,
					     {'onlybody' => 1}));
    if ((keys %msgs)[0]=~/^error\:/) {
	$r->print(&mt('No stored comments yet.'));
    } else {
	my $found=0;
	foreach my $key (sort(keys(%msgs))) {
	    $r->print("\n".$msgs{$key}."<hr />");
	    $found=1;
	}
	unless ($found) {
	    $r->print(&mt('No stored comments yet for this resource.'));
	}
    }
}

# ---------------------------------------------------------------- Send an email

sub sendoffmail {
    my ($r,$folder)=@_;
    my $suffix=&Apache::lonmsg::foldersuffix($folder);
    my $sendstatus='';
    my %specialmsg_status;
    my $numspecial = 0;
    my ($cdom,$cnum,$group);
    if (exists($env{'form.group'})) {
        $group = $env{'form.group'};
    }
    if (exists($env{'request.course.id'})) {
        $cnum = $env{'course.'.$env{'request.course.id'}.'.num'};
        $cdom = $env{'course.'.$env{'request.course.id'}.'.domain'};
    }
    if ($env{'form.send'}) {
        if ($group eq '') {
	    &printheader($r,'','Messages being sent.');
        } else {
            $r->print(&groupmail_header('sending',$group));
        }
	$r->rflush();
	my %content=();
	undef %content;
	if ($env{'form.forwid'}) {
	    my $msgid=$env{'form.forwid'};
	    my %message=&Apache::lonnet::get('nohist_email'.$suffix,[$msgid]);
	    %content=&Apache::lonmsg::unpackagemsg($message{$msgid},1);
	    &statuschange($msgid,'forwarded',$folder);
	    $env{'form.message'}.="\n\n-- Forwarded message --\n\n".
		$content{'message'};
	}
	if ($env{'form.replyid'}) {
	    my $msgid=$env{'form.replyid'};
	    my %message=&Apache::lonnet::get('nohist_email'.$suffix,[$msgid]);
	    %content=&Apache::lonmsg::unpackagemsg($message{$msgid},1);
	    &statuschange($msgid,'replied',$folder);
	}

	my @to =
	    &Apache::loncommon::get_env_multiple('form.selectedusers_forminput');
	my $mode = $env{'form.sendmode'};

	my %toaddr;
	if (@to) {
	    foreach my $dest (@to) {
		my ($user,$domain) = split(/:/, $dest);
		if (($user ne '') && ($domain ne '')) {
		    my $address = $user.":".$domain; # How the code below expects it.
		    $toaddr{$address} = '';
		}
	    }
	}

	if ($env{'form.sendmode'} eq 'group') {
	     foreach my $address (keys(%env)) {
	 	if ($address=~/^form\.send\_to\_\&\&\&[^\&]*\&\&\&\_(.+)$/) {
	 	    $toaddr{$1}='';
	 	}
	    }
	} elsif ($env{'form.sendmode'} eq 'upload') {
	    foreach my $line (split(/[\n\r\f]+/,$env{'form.upfile'})) {
                my ($rec,$txt) = ($line =~ /^([^:]+:[^:]+):(.*)$/);
		if ($txt) {
                    $rec =~ s/^\s+//;
                    $rec =~ s/\s+$//;
		    $toaddr{$rec}.=$txt."\n";
		}
	    }
	} else {
	    if (($env{'form.recuname'} ne '') && ($env{'form.recdomain'} ne '')) {
		$toaddr{$env{'form.recuname'}.':'.$env{'form.recdomain'}}='';
	    }
	}
	if ($env{'form.additionalrec'}) {
	    foreach my $rec (split(/\,/,$env{'form.additionalrec'})) {
		my ($auname,$audom)=split(/:/,$rec);
		if (($auname ne "") && ($audom ne "")) {
		    $toaddr{$auname.':'.$audom}='';
		}
	    }
	}

        my $savemsg;
        my $msgtype;
        my %sentmessage;
        my $msgsubj=&Apache::lonfeedback::clear_out_html($env{'form.subject'},
							 undef,1);
        if ((($env{'form.critmsg'}) || ($env{'form.sendbck'})) &&
            (&Apache::lonnet::allowed('srm',$env{'request.course.id'})
	     || &Apache::lonnet::allowed('srm',$env{'request.course.id'}.
					 '/'.$env{'request.course.sec'})
	     )) {
            $savemsg=&Apache::lonfeedback::clear_out_html($env{'form.message'},1);
            $msgtype = 'critical';
        } else {
            $savemsg=&Apache::lonfeedback::clear_out_html($env{'form.message'});
        }
	
	foreach my $address (sort(keys(%toaddr))) {
	    my ($recuname,$recdomain)=split(/\:/,$address);
            my $msgtxt = $savemsg;
	    if ($toaddr{$address}) { $msgtxt.='<hr />'.$toaddr{$address}; }
	    my @thismsg;
	    if ((($env{'form.critmsg'}) || ($env{'form.sendbck'})) && 
		(&Apache::lonnet::allowed('srm',$env{'request.course.id'})
		 || &Apache::lonnet::allowed('srm',$env{'request.course.id'}.
					     '/'.$env{'request.course.sec'}))) {
		$r->print(&mt('Sending critical message').' '.$recuname.':'.$recdomain.': ');
		@thismsg=
		    &Apache::lonmsg::user_crit_msg($recuname,$recdomain,
						   $msgsubj,$msgtxt,
						   $env{'form.sendbck'},
						   $env{'form.permanent'},
						   \$sentmessage{$address});
	    } else {
		$r->print(&mt('Sending').' '.$recuname.':'.$recdomain.': ');
		@thismsg=
		    &Apache::lonmsg::user_normal_msg($recuname,$recdomain,
						     $msgsubj,$msgtxt,
						     $content{'citation'},
						     undef,undef,
						     $env{'form.permanent'},
						     \$sentmessage{$address});
            }
	    if (($env{'request.course.id'}) && (($msgtype eq 'critical') || 
                                         ($env{'form.sendmode'} eq 'group'))) {
	        $specialmsg_status{$recuname.':'.$recdomain} =
		    join(' ',@thismsg);
		foreach my $result (@thismsg) {
		    if ($result eq 'ok') {
			$numspecial++;
		    }
		}
	    }
	    $sendstatus.=' '.join(' ',@thismsg);
	}
        if (($env{'request.course.id'}) && (($env{'form.sendmode'} eq 'group')
                                              || ($msgtype eq 'critical'))) {
            my $subj_prefix;
            if ($msgtype eq 'critical') {
                $subj_prefix = 'Critical.';
            } else {
                $subj_prefix = 'Broadcast.';
            }
            my ($specialmsgid,$specialresult);
            my $course_str = &escape('['.$cnum.':'.$cdom.']');

            if ($numspecial) {
                $specialresult = &Apache::lonmsg::user_normal_msg_raw($cnum,$cdom,$subj_prefix.
                    ' '.$course_str,$savemsg,undef,undef,undef,
                    undef,undef,\$specialmsgid);
                $specialmsgid = &unescape($specialmsgid);
            }
            if ($specialresult eq 'ok') {
                my $record_sent;
                my @recusers;
                my @recudoms;
                my ($stamp,$crssubj,$msgname,$msgdom,$msgcount,$context,$pid) =
		    split(/\:/,&unescape($specialmsgid));

                foreach my $recipient (sort(keys(%toaddr))) {
                    if ($specialmsg_status{$recipient} eq 'ok') {
                        my $usersubj = $subj_prefix.'['.$recipient.']';
                        my $usermsgid = 
			    &Apache::lonmsg::buildmsgid($stamp,$usersubj,
							$msgname,$msgdom,
							$msgcount,$context,
							$pid);
                        &Apache::lonmsg::user_normal_msg_raw($cnum,$cdom,$subj_prefix.
                                             ' ['.$recipient.']',$msgsubj,undef,
                        undef,undef,undef,$usermsgid,undef,undef,$specialmsgid);
                        my ($uname,$udom) = split(/:/,$recipient);
                        push(@recusers,$uname);
                        push(@recudoms,$udom);
                    }
                }
                if (@recusers) {
                    my $specialmessage;
                    my $sentsubj = 
			$subj_prefix.' ('.$numspecial.' sent) '.$msgsubj;
                    $sentsubj = &HTML::Entities::encode($sentsubj,'<>&"');
                    my $sentmsgid = 
			&Apache::lonmsg::buildmsgid($stamp,$sentsubj,$msgname,
						    $msgdom,$msgcount,$context,
						    $pid);
                    ($specialmsgid,$specialmessage) = &Apache::lonmsg::packagemsg($msgsubj,$savemsg,
                            undef,undef,undef,\@recusers,\@recudoms,$sentmsgid);
                    $record_sent = &Apache::lonmsg::store_sent_mail($specialmsgid,$specialmessage);
                }
            } else {
                &Apache::lonnet::logthis('Failed to create record of critical message or broadcast in '.$env{'course.'.$env{'request.course.id'}.'.num'}.' at '.$env{'course.'.$env{'request.course.id'}.'.domain'}.' - no msgid generated');
            }
        }
    } else {
	&printheader($r,'','No messages sent.'); 
    }
    if ($sendstatus=~/^(\s*(?:ok|con_delayed)\s*)*$/) {
	$r->print('<br /><span class="LC_success">'.&mt('Completed.').'</span>');
	if ($env{'form.displayedcrit'}) {
	    &discrit($r);
        }
        if ($group ne '') {
            $r->print(&groupmail_sent($group,$cdom,$cnum)); 
	} else {
	    &Apache::loncommunicate::menu($r);
	}
    } else {
	$r->print('<p><span class="LC_error">'.&mt('Could not deliver message').'</span> '.
		  &mt('Please use the browser "Back" button and correct the recipient addresses '."($sendstatus)").'</p>');
    }
}

# ===================================================================== Handler

sub handler {
    my $r=shift;

# ----------------------------------------------------------- Set document type
    
    &Apache::loncommon::content_type($r,'text/html');
    $r->send_http_header;
    
    return OK if $r->header_only;
    
# --------------------------- Get query string for limited number of parameters
    &Apache::loncommon::get_unprocessed_cgi($ENV{'QUERY_STRING'},
        ['display','replyto','forward','markread','markdel','markunread',
         'sendreply','compose','sendmail','critical','recname','recdom',
         'recordftf','sortedby','block','folder','startdis','interdis',
	 'showcommentbaseurl','dismode','group','subject','text','ref']);
    $sqs='&sortedby='.$env{'form.sortedby'};

# ------------------------------------------------------ They checked for email
    unless ($env{'form.block'}) {
        &Apache::lonnet::put('email_status',{'recnewemail'=>0});
    }

# ----------------------------------------------------------------- Breadcrumbs

    &Apache::lonhtmlcommon::clear_breadcrumbs();
    &Apache::lonhtmlcommon::add_breadcrumb
        ({href=>"/adm/communicate",
          text=>"Communication/Messages",
          faq=>12,bug=>'Communication Tools',});

# ------------------------------------------------------------------ Get Folder

    my $folder=$env{'form.folder'};
    unless ($folder) { 
	$folder=''; 
    } else {
	$sqs.='&folder='.&escape($folder);
    }
# ------------------------------------------------------------ Get Display Mode

    my $dismode=$env{'form.dismode'};
    unless ($dismode) { 
	$dismode=''; 
    } else {
	$sqs.='&amp;dismode='.&escape($dismode);
    }

# --------------------------------------------------------------------- Display

    $startdis=$env{'form.startdis'};
    $startdis--;
    unless ($startdis) { $startdis=0; }

    $interdis=$env{'form.interdis'};
    unless ($interdis) { $interdis=20; }
    $sqs.='&interdis='.$interdis;

    if ($env{'form.firstview'}) {
	$startdis=0;
    }
    if ($env{'form.lastview'}) {
	$startdis=-1;
    }
    if ($env{'form.prevview'}) {
	$startdis--;
    }
    if ($env{'form.nextview'}) {
	$startdis++;
    }
    my $postedstartdis=$startdis+1;
    $sqs.='&startdis='.$postedstartdis;

# --------------------------------------------------------------- Render Output

    if ($env{'form.display'}) {
	&displaymessage($r,$env{'form.display'},$folder);
    } elsif ($env{'form.replyto'}) {
	&compout($r,'',$env{'form.replyto'},undef,undef,$folder,$dismode);
    } elsif ($env{'form.confirm'}) {
	&printheader($r,'','Confirmed Receipt');
	my $replying = 0;
	foreach my $envkey (keys(%env)) {
	    if ($envkey=~/^form\.rec\_(.*)$/) {
		$r->print('<b>'.&mt('Confirming Receipt').':</b> '.
			  &Apache::lonmsg::user_crit_received($1).'<br>');
	    }
	    if ($envkey=~/^form\.reprec\_(.*)$/) {
		my $msgid=$1;
		$r->print('<b>'.&mt('Confirming Receipt').':</b> '.
			  &Apache::lonmsg::user_crit_received($msgid).'<br>');
		&compout($r,'','','',$msgid);
		$replying = 1;
	    }
	}
	if (!$replying) {
	    &discrit($r);
	}
    } elsif ($env{'form.critical'}) {
	&printheader($r,'','Displaying Critical Messages');
	&discrit($r);
    } elsif ($env{'form.forward'}) {
	&compout($r,$env{'form.forward'},undef,undef,undef,$folder);
    } elsif ($env{'form.markdel'}) {
	&printheader($r,'','Deleted Message');
	my ($result,$msg) = 
	    &statuschange($env{'form.markdel'},'deleted',$folder);
	if (!$result) {
	    $r->print('<p class="LC_error">'.
		      &mt('Failed to delete the message.').'</p>'.
		      '<p class="LC_error">'.$msg."</p>\n");
	}
	&Apache::loncommunicate::menu($r);
	&disall($r,($folder?$folder:$dismode));
    } elsif ($env{'form.markedmove'}) {
	my ($total,$failed,@failed_msg)=(0,0);
	foreach my $key (keys(%env)) {
	    if ($key=~/^form\.delmark_(.*)$/) {
		my ($result,$msg) =
		    &movemsg(&unescape($1),$folder,
			     $env{'form.movetofolder'});
		if ($result) {
		    $total++;
		} else {
		    $failed++;
		    push(@failed_msg,$msg);
		}
	    }
	}
	&printheader($r,'','Moved Messages');
	if ($failed) {
	    $r->print('<p class="LC_error">
                          '.&mt('Failed to move [_1] message(s)',$failed).
		      '</p>');
	    $r->print('<p class="LC_error">'.
		      join("</p>\n<p class=\"LC_error\">",@failed_msg).
		      "</p>\n");
	}
	$r->print(&mt('Moved [_1] message(s)',$total).'<p>');
	&Apache::loncommunicate::menu($r);
	&disall($r,($folder?$folder:$dismode));
    } elsif ($env{'form.markeddel'}) {
	my ($total,$failed,@failed_msg)=(0,0);
	foreach my $key (keys(%env)) {
	    if ($key=~/^form\.delmark_(.*)$/) {
		my ($result,$msg) = 
		    &statuschange(&unescape($1),'deleted',
				  $folder);
		if ($result) {
		    $total++;
		} else {
		    $failed++;
		    push(@failed_msg,$msg);
		}
	    }
	}
	&printheader($r,'','Deleted Messages');
	if ($failed) {
	    $r->print('<p class="LC_error">
                          '.&mt('Failed to delete [_1] message(s)',$failed).
		      '</p>');
	    $r->print('<p class="LC_error">'.
		      join("</p>\n<p class=\"LC_error\">",@failed_msg).
		      "</p>\n");
	}
	$r->print(&mt('Deleted [_1] message(s)',$total).'<p>');
	&Apache::loncommunicate::menu($r);
	&disall($r,($folder?$folder:$dismode));
    } elsif ($env{'form.markunread'}) {
	&printheader($r,'','Marked Message as Unread');
	&statuschange($env{'form.markunread'},'new');
	&Apache::loncommunicate::menu($r);
	&disall($r,($folder?$folder:$dismode));
    } elsif ($env{'form.compose'}) {
	&compout($r,'','',$env{'form.compose'});
    } elsif ($env{'form.recordftf'}) {
	&facetoface($r,$env{'form.recordftf'});
    } elsif ($env{'form.block'}) {
        &examblock($r,$env{'form.block'});
    } elsif ($env{'form.sendmail'}) {
	&sendoffmail($r,$folder);
	if ($env{'form.storebasecomment'}) {
	    &storecomment($r);
        }
	if (($env{'form.rsspost'}) && ($env{'request.course.id'})) {
	        &Apache::lonrss::addentry($env{'course.'.$env{'request.course.id'}.'.num'},
				      $env{'course.'.$env{'request.course.id'}.'.domain'},
				      'Course_Announcements',
				      $env{'form.subject'},
				      $env{'form.message'},'/adm/communicate','public');
	}
	if ((!exists($env{'form.group'})) && (!$env{'form.displayedcrit'})) {
	    &disall($r,($folder?$folder:$dismode));
	}
    } elsif ($env{'form.newfolder'}) {
	&printheader($r,'','New Folder');
        my $showfolder = $env{'form.newfolder'};
	my ($makeresult,$warning) = &makefolder($env{'form.newfolder'});
        if ($makeresult eq 'ok') {
            $r->print(&mt('Mail folder "[_1]" created.',$showfolder).'<br />');
        } else {
            $r->print(&mt('Creation failed.').' '.$makeresult.'<br />'.
                      $warning);
            $showfolder = $folder;
        }
        &Apache::loncommunicate::menu($r);
	&disall($r,$showfolder);
    } elsif ($env{'form.showcommentbaseurl'}) {
	&storedcommentlisting($r);
    } elsif ($env{'form.folderaction'} eq 'delete') {
        &printheader($r,'','Deleted Folder');
        my $showfolder = '';
        my $delresult = &deletefolder($folder);
        if ($delresult eq 'ok') {
            $r->print(&mt('Mail folder "[_1]" deleted.',$folder).'<br />');
        } else {
            $r->print(&mt('Deletion failed.').' '.$delresult.'<br />');
            $showfolder = $folder;
        }
        &Apache::loncommunicate::menu($r);
        &disall($r,$showfolder);
    } elsif ($env{'form.folderaction'} eq 'rename') {
        &printheader($r,'','Renamed Folder');
        my $showfolder = $env{'form.renamed'};
        my $renresult = &renamefolder($folder);
        if ($renresult eq 'ok') {
            $r->print(&mt('Mail folder "[_1]" renamed "[_2]".',$folder,$showfolder).'<br />');
        } else {
            $r->print(&mt('Renaming failed.').' '.$renresult.'<br />');
            $showfolder = $folder;
        }
        &Apache::loncommunicate::menu($r);
        &disall($r,$showfolder);
    } else {
	&printheader($r,'','Display All Messages');
	&Apache::loncommunicate::menu($r);
	&disall($r,($folder?$folder:$dismode));
    }
    $r->print(&Apache::loncommon::end_page());
    return OK;
}
# ================================================= Main program, reset counter

=pod

=back

=cut

1; 

__END__

